/*	Style Implementation for Hypertext			HTNSStyle.c
**	==================================
**
**	Styles allow the translation between a logical property
**	of a piece of text and its physical representation.
**
**	A StyleSheet is a collection of styles, defining the
**	translation necessary to
**	represent a document. It is a linked list of styles.
*/
/*	Style Module for NextStep-based applications for WWW
*/

#import "HTNSStyle.h"		/* implemented by this module */

#import "libwww.h"

extern HTStyleSheet * styleSheet;	/* Application-wide */

/*	Module-wide style cache
*/
PRIVATE int 		got_styles = 0;
PRIVATE HTStyle *styles[HTMLP_ELEMENTS];
PRIVATE HTStyle *default_style;

/*	Get Styles from style sheet
**	---------------------------
*/
PRIVATE void get_styles NOARGS
{
    got_styles = YES;
    
    default_style =		HTStyleNamed(styleSheet, "Normal");

    styles[HTML_H1] =		HTStyleNamed(styleSheet, "Heading1");
    styles[HTML_H2] =		HTStyleNamed(styleSheet, "Heading2");
    styles[HTML_H3] =		HTStyleNamed(styleSheet, "Heading3");
    styles[HTML_H4] =		HTStyleNamed(styleSheet, "Heading4");
    styles[HTML_H5] =		HTStyleNamed(styleSheet, "Heading5");
    styles[HTML_H6] =		HTStyleNamed(styleSheet, "Heading6");
    styles[HTML_H7] =		HTStyleNamed(styleSheet, "Heading7");

    styles[HTML_DL] =		HTStyleNamed(styleSheet, "Glossary");
    styles[HTML_UL] =
    styles[HTML_OL] =		HTStyleNamed(styleSheet, "List");
    styles[HTML_MENU] =		HTStyleNamed(styleSheet, "Menu");
    styles[HTML_DIR] =		HTStyleNamed(styleSheet, "Dir");    
/*  styles[HTML_DLC] =		HTStyleNamed(styleSheet, "GlossaryCompact"); */
    styles[HTML_ADDRESS]=	HTStyleNamed(styleSheet, "Address");
    styles[HTML_BLOCKQUOTE]=	HTStyleNamed(styleSheet, "BlockQuote");
    styles[HTML_PLAINTEXT] =
    styles[HTML_XMP] =		HTStyleNamed(styleSheet, "Example");
    styles[HTML_PRE] =		HTStyleNamed(styleSheet, "Preformatted");
    styles[HTML_LISTING] =	HTStyleNamed(styleSheet, "Listing");
}


/*	Generate the style for a new combination
**	----------------------------------------
**
**	This SHOULD use the style sheet and allow incremental
**	changes for nesting things, and dependencies on attribute
**	values :-( .  Simple..
*/

#define BLOCKQUOTE_INDENT	20
#define GLOSSARY_INDENT 	150
#define LIST_INDENT 		30

/*	Set a hangling indent and first tab stop
**	----------------------------------------
*/
PRIVATE void set_hanging_indent ARGS2(HTStyle *, style, int, i)
{
    NXTextStyle *p = &(style->paragraph);
    if (p->numTabs>0 && p->tabs) free(p->tabs);

    p->numTabs = 1;
    p->tabs = malloc(1 * sizeof(p->tabs[0]));
    p->tabs[0].kind = 0;
    p->tabs[0].x = p->indent1st + i;
    p->indent2nd = p->indent1st + i;
}


/*	This is a PLATFORM DEPENDENT ROUTINE which should be in the
**	style module
**
**	The sort of style sheet we really want is something which
**	will contain an interpreter whose language allowws expression
**	of anything we may want to put in here.
*/
PRIVATE FontManager * fm = NULL;
PRIVATE HTStyle * set_trait(HTStyle* style, NXFontTraitMask traits)
{
    style->font = [fm convert:style->font toHaveTrait: traits];
    return style;
    
}

/*	This sets the style associated with the new nesting
**	level.  It also sets the paragraph_break flag if necessary.
**	(It is initially zero in new nesting)
*/
PUBLIC HTStyle * HTStyleModify(HTStyle * old, 
	HTNesting* nesting, int element_number)
{
    HTStyle* style = malloc(sizeof(HTStyle));
    if (!style) outofmem(__FILE__, "new style");
    
    if (!old) old = HTStyleNamed(styleSheet, "Normal");

    *style = *old;		/* Default to style of enclosing element */

    style->nesting = nesting;	/* Link back */
    
    if (style->paragraph.numTabs > 0) {
        int tab;
	style->paragraph.tabs = malloc(
		style->paragraph.numTabs *
			sizeof(style->paragraph.tabs[0]));
	for (tab=0; tab < style->paragraph.numTabs; tab++) {
	    style->paragraph.tabs[tab] = 
	    	old->paragraph.tabs[tab];
	}
    }
    
    if (element_number<0) return style;  /* No modification */
    
    if (!fm) fm = [FontManager new];
    switch(element_number) {
    
        case HTML_HR: 		nesting->paragraph_break = YES;
				nesting->object_gens_HTML = YES;
				return style;
	case HTML_IMG:
	case HTML_INPUT:	nesting->object_gens_HTML = YES;
				break;
	
        case HTML_A:            style->textRGBColor = 0x007700;
                                break;
	case HTML_B:		return set_trait(style, NX_BOLD);
	case HTML_I:		return set_trait(style, NX_ITALIC);
	case HTML_STRONG:	return set_trait(style, NX_BOLD);
	case HTML_CITE:		return set_trait(style, NX_ITALIC);
	case HTML_U:		style->underline = YES;
				return style; 
	case HTML_EM:		return set_trait(style, NX_BOLD);
	case HTML_TT:		style->font = [fm convert:style->font
					toFamily:"Courier"];
				return style;

//	The P is a little strange.  It doesn't have to have a paragraph
//	around it as it is a paragraph break.  So the flag isn't set
//	here or we get a double break. Compare with HR which has an
//	image and must be surrounded by paragraph break.
//	Also, we specify that the HTML generation is left to P-specific code.

	case HTML_P:		// nesting->wst = WS_NORMAL;
				// nesting->paragraph_break = YES;
				nesting->object_gens_HTML = YES;    // above
				return style;	/* Doesn't change style */



	case HTML_BLOCKQUOTE:	nesting->wst = WS_NORMAL;
				nesting->paragraph_break = YES;
				style->paragraph.indent1st =
				    style->paragraph.indent1st +
				     BLOCKQUOTE_INDENT;
				style->paragraph.indent2nd =
				    style->paragraph.indent2nd +
				     BLOCKQUOTE_INDENT;
				set_trait(style, NX_ITALIC);
				break;
				
	case HTML_ADDRESS:	nesting->wst = WS_NORMAL;
				nesting->paragraph_break = YES;
				style->paragraph.alignment = NX_RIGHTALIGNED;
				return set_trait(style, NX_ITALIC);

	case HTML_H1:
	case HTML_H2:
	case HTML_H3:
	case HTML_H4:
	case HTML_H5:
	case HTML_H6:		nesting->wst = WS_HEADING;
				nesting->paragraph_break = YES;
				break;
	case HTML_XMP:
	case HTML_LISTING:
	case HTML_PLAINTEXT:
	case HTML_PRE:		nesting->wst = WS_PRE;
				nesting->paragraph_break = YES;
				break;
	case HTML_UL:
	case HTML_OL:
	case HTML_MENU:		nesting->item_number = 0;
				style->paragraph.indent1st =
				    style->paragraph.indent2nd;

				set_hanging_indent(
				    style,
				    LIST_INDENT);
				nesting->list_level++;

				nesting->wst = WS_LIST;
				nesting->paragraph_break = YES;
				return style;  /* style is fixed up OK */
				
	case HTML_DL:
				nesting->item_number = 0;
				style->paragraph.indent1st =
				    style->paragraph.indent2nd;

				set_hanging_indent(
				    style,
				    GLOSSARY_INDENT);
				nesting->list_level++;

				nesting->wst = WS_GLOSSARY;
				nesting->paragraph_break = YES;
				return style;  /* style is fixed up OK */
				
	case HTML_DIR:		nesting->wst = WS_DIR;
				nesting->paragraph_break = YES;
				break;

	case HTML_FORM:		style->textGray = 0.667;
				style->textRGBColor = 0x770000;
				nesting->paragraph_break = YES;
				break;
				
	default:

				break;
    }  /* switch on element */
    
	/* @@@@@@@ placeholder only --- will go! @@@  */
    {
	HTStyle * s;
	if (!got_styles) get_styles();
	s = styles[element_number];
	if (s) *style = *s;
	style->nesting = nesting;	/* Link back */
    }
    
    return style;
}
/*////////////////////////////////////////////////////////////////////
*/

/*	Create a new style
*/
PUBLIC HTStyle* HTStyleNew NOARGS
{
    return (HTStyle *)calloc(1, sizeof(HTStyle));
}

/*	Create a new style with a name
*/
PUBLIC HTStyle* HTStyleNewNamed ARGS1 (CONST char *,name)
{
    HTStyle * self = HTStyleNew();
    StrAllocCopy(self->name, name);
    return self;
}


/*	Free a style
*/
PUBLIC HTStyle * HTStyleFree ARGS1 (HTStyle *,self)
{
    if (self->name) free(self->name);
    if (self->SGMLTag) free(self->SGMLTag);
    free(self);
    return 0;
}


/*	Read a style from a stream	(without its name)
**	--------------------------
**
**	Reads a style with paragraph information from a stream.
**	The style name is not read or written by these routines.
*/
#define NONE_STRING "(None)"
HTStyle * HTStyleRead(HTStyle * style, NXStream * stream)
{
    char myTag[STYLE_NAME_LENGTH];
    char fontName[STYLE_NAME_LENGTH];
    int	tab;
    int gotpara;		/* flag: have we got a paragraph definition? */

    style->magic = HTSTYLE_MAGIC_NUMBER;	
    NXScanf(stream, "%s%i%s%f%i",
	myTag,
	&style->SGMLType,
	fontName,
	&style->fontSize,
	&gotpara);
    if (gotpara) {
	NXTextStyle *p = &(style->paragraph);
	if (p->tabs) free(p->tabs);
	p->tabs = 0;
	NXScanf(stream, "%f%f%f%f%hd%f%f%hd",
	    &p->indent1st,
	    &p->indent2nd,
	    &p->lineHt,
	    &p->descentLine,
	    &p->alignment,
	    &style->spaceBefore,
	    &style->spaceAfter,
	    &p->numTabs);
	p->tabs = malloc(p->numTabs * sizeof(p->tabs[0]));
	for (tab=0; tab < p->numTabs; tab++) {
	    NXScanf(stream, "%hd%f",
		    &p->tabs[tab].kind,
		    &p->tabs[tab].x);
	}
    } else { /* No paragraph */
    	    style->paragraph.lineHt = 0;
    } /* if no paragraph */
    StrAllocCopy(style->SGMLTag, myTag);
    if (strcmp(fontName, NONE_STRING)==0)
        style->font = 0;
    else
        style->font = [Font newFont:fontName size:style->fontSize];
    return 0;
}


/*	Write a style to a stream in a compatible way
*/
HTStyle * HTStyleWrite(HTStyle * style, NXStream * stream)
{
    int tab;
    NXTextStyle *p = &style->paragraph;
    NXPrintf(stream, "%s %i %s %f %i\n",
	style->SGMLTag,
	style->SGMLType,
	style->font ? [style->font name] : NONE_STRING,
	style->fontSize,
	p!=0);

    if (p) {
	NXPrintf(stream, "\t%f %f %f %f %i %f %f\t%i\n",
	    p->indent1st,
	    p->indent2nd,
	    p->lineHt,
	    p->descentLine,
	    p->alignment,
	    style->spaceBefore,
	    style->spaceAfter,
	    p->numTabs);
	    
	for (tab=0; tab < p->numTabs; tab++)
	    NXPrintf(stream, "\t%i %f\n",
		    p->tabs[tab].kind,
		    p->tabs[tab].x);
	}
    return style;
}


/*	Write a style to stdout for diagnostics
*/
HTStyle * HTStyleDump(HTStyle * style)
{
    int tab;
    NXTextStyle *p = &style->paragraph;
    printf("Style %p `%s' SGML:%s, type=%i. Font %s %.1f point.\n",
    	(void*)style,
	style->name,
	style->SGMLTag,
	style->SGMLType,
	[style->font name],
	style->fontSize);
    if (p) {
        printf(
    	"\tIndents: first=%.0f others=%.0f, Height=%.1f Desc=%.1f\n"
	"\tAlign=%i, %i tabs. (%.0f before, %.0f after)\n",
	    p->indent1st,
	    p->indent2nd,
	    p->lineHt,
	    p->descentLine,
	    p->alignment,
	    p->numTabs,
	    style->spaceBefore,
	    style->spaceAfter);
	    
	for (tab=0; tab < p->numTabs; tab++) {
	    printf("\t\tTab kind=%i at %.0f\n",
		    p->tabs[tab].kind,
		    p->tabs[tab].x);
    	}
	printf("\n");
    } /* if paragraph */
    return style;
}



/*			StyleSheet Functions
**			====================
*/

/*	Searching for styles:
*/
HTStyle * HTStyleNamed ARGS2 (HTStyleSheet *,self, CONST char *,name)
{
    HTStyle * scan;
    for (scan=self->styles; scan; scan=scan->next)
        if (0==strcmp(scan->name, name)) return scan;
    if (TRACE) fprintf(stderr, "StyleSheet: No style named `%s'\n", name);
    return 0;
}

HTStyle * HTStyleForParagraph(HTStyleSheet * self, NXTextStyle *para)
{
    HTStyle * s = (HTStyle*) para;
    if (s->magic == HTSTYLE_MAGIC_NUMBER)
        return s;
    else return NULL;
}

/*	Find the style which best fits a given run
**	------------------------------------------
**
**	This heuristic is used for guessing the style for a run of
**	text which has been pasted in. In order, we try:
**
**	A style whose paragraph structure is actually used by the run.
**	A style matching in font
**	A style matching in paragraph style exactly
**	A style matching in paragraph to a degree
**	************** Doesn't return a valid nesting! 
*/
HTStyle * HTStyleForRun(HTStyleSheet *self, NXRun *run)
{
    HTStyle * scan;
    HTStyle * best = 0;
    int	bestMatch = 0;
    NXTextStyle * rp = run->paraStyle;
    scan = (HTStyle *) run->paraStyle;
    if (scan && (scan->magic == HTSTYLE_MAGIC_NUMBER)) return scan;
    
    for (scan=self->styles; scan; scan=scan->next)
        if (&scan->paragraph == run->paraStyle) return scan;	/* Exact */

    for (scan=self->styles; scan; scan=scan->next){
    	NXTextStyle * sp = &scan->paragraph;
    	if (sp->lineHt) {
	    int match = 0;
	    if (sp->indent1st ==	rp->indent1st)	match = match+1;
	    if (sp->indent2nd ==	rp->indent2nd)	match = match+2;
	    if (sp->lineHt ==		rp->lineHt)	match = match+1;
	    if (sp->numTabs ==		rp->numTabs)	match = match+1;
	    if (sp->alignment ==	rp->alignment)	match = match+3;
	    if (scan->font ==		run->font)	match = match+10;
	    if (match>bestMatch) {
		    best=scan;
		    bestMatch=match;
	    }
	}
    }
    if (TRACE) printf("HTStyleForRun: Best match for style is %i out of 18\n",
    			 bestMatch);
    return best;
}




/*	Add a style to a sheet
**	----------------------
*/
HTStyleSheet * HTStyleSheetAddStyle ARGS2
  (HTStyleSheet *,self, HTStyle *,style)
{
    style->next = 0;		/* The style will go on the end */
    if (!self->styles) {
    	self->styles = style;
    } else {
    	HTStyle * scan;
        for(scan=self->styles; scan->next; scan=scan->next); /* Find end */
	scan->next=style;
    }
    return self;
}


/*	Remove the given object from a style sheet if it exists
*/
HTStyleSheet * HTStyleSheetRemoveStyle ARGS2
  (HTStyleSheet *,self, HTStyle *,style)
{
    if (self->styles == style) {
    	self->styles = style->next;
	return self;
    } else {
    	HTStyle * scan;
	for(scan = self->styles; scan; scan = scan->next) {
	    if (scan->next == style) {
	        scan->next = style->next;
		return self;
	    }
	}
    }
    return 0;
}

/*	Create new style sheet
*/

HTStyleSheet * HTStyleSheetNew NOARGS
{
    HTStyleSheet * self = (HTStyleSheet *)malloc(sizeof(*self));

    memset((void*)self, 0, sizeof(*self));	/* ANSI */
/* Harbison c ref man says (char*)self
   but k&r ansii and abc books and Think_C say (void*) */
    
/*    bzero(self, sizeof(*self)); */		/* BSD */
    return self;
}


/*	Free off a style sheet pointer
*/
HTStyleSheet * HTStyleSheetFree ARGS1 (HTStyleSheet *,self)
{
    HTStyle * style;
    while((style=self->styles)!=0) {
        self->styles = style->next;
	HTStyleFree(style);
    }
    free(self);
    return 0;
}


/*	Read a stylesheet from a typed stream
**	-------------------------------------
**
**	Reads a style sheet from a stream.  If new styles have the same names
**	as existing styles, they replace the old ones without changing the ids.
*/

HTStyleSheet * HTStyleSheetRead(HTStyleSheet * self, NXStream * stream)
{
    int numStyles;
    int i;
    HTStyle * style;
    char styleName[80];
    NXScanf(stream, " %d ", &numStyles);
    if (TRACE) fprintf(stderr, "Stylesheet: Reading %d styles\n", numStyles);
    for (i=0; i<numStyles; i++) {
        NXScanf(stream, "%s", styleName);
        style = HTStyleNamed(self, styleName);
	if (!style) {
	    style = HTStyleNewNamed(styleName);
	    (void) HTStyleSheetAddStyle(self, style);
	}
	(void) HTStyleRead(style, stream);
	if (TRACE) HTStyleDump(style);
    }
    return self;
}

/*	Write a stylesheet to a typed stream
**	------------------------------------
**
**	Writes a style sheet to a stream.
*/

HTStyleSheet * HTStyleSheetWrite(HTStyleSheet * self, NXStream * stream)
{
    int numStyles = 0;
    HTStyle * style;
    
    for(style=self->styles; style; style=style->next) numStyles++;
    NXPrintf(stream, "%d\n", numStyles);
    
    if (TRACE) fprintf(stderr, "StyleSheet: Writing %d styles\n", numStyles);
    for (style=self->styles; style; style=style->next) {
        NXPrintf(stream, "%s ", style->name);
	(void) HTStyleWrite(style, stream);
    }
    return self;
}


