//	HyperText Implementation				    HyperText.m
//	------------------------
//
// HyperText is like Text, but includes links to and from other hypertexts.
//
// Authors:
//	TBL		Tim Berners-Lee CERN/CN
//
// See also:
//	The HTAnchor class, and the HyperAccess class.
//
// History:
//	25 Sep 90	Written (TBL) with the  help of the Interface builder.
//	14 Mar 91	Page width is taken from application's page layout.


// Notes.
//
//	For all anchors to have addresses, the node must be made with
//	newAnchor:Server: . Do not use any other creation methods inherited.

//#import <appkit/appkit.h>
#import "HyperText.h"

#import "HyperManager.h"
//#import "HTUtils.h"
//#import "HTParse.h"
#import "HTNSStyle.h"		/* Need this one */
//#import "HTFormat.h"
#import "HyperCell.h"
#import "ImageCell.h"
//#import "HTAccess.h"
//#include "HTML.h"		/* For directory object building */
//#include "SGML.h"		/* Structured stream def */



@implementation HyperText



struct _HTStructured {
	HTStructuredClass * isa;
	/* ... */
};

typedef struct {
	BOOL		insert;
	int		level;
	int		element_number;
	BOOL		set_info;
	HTChildAnchor *	info;
} HTOperation;


PUBLIC BOOL HTSuppressImages = NO;

#define DEFAULT_TITLE "World-Wide Web Document"

#define ANCHOR_ID_PREFIX "z"	/* Auto anchors run z1, z2, ... 921122 */

extern        HTStyleSheet * styleSheet;	/* see StyleToy */

static int window_sequence = 0;		/* For stacking windows neatly */
#define SLOTS	30			/* Number of position slots */
static HyperText * slot[SLOTS];			/* Ids of HT objects taking them */

PUBLIC BOOL HTMultiWindowMode = NO;

#define NICE_HEIGHT 600.0	/* Allows a few windows to be stacked */
#define MAX_HEIGHT 720.0	/* Worth going bigger to get it all in	*/
#define MIN_HEIGHT 80.0		/* Mustn't lose the window! */
#define MIN_WIDTH 200.0

+initialize
{
    int i;
    for (i=0; i<SLOTS; i++) slot[i] = 0;
    return [super initialize];
}

//	Get Application page layout's Page Width
//	----------------------------------------
//
//	Returned in pixels
//
PRIVATE float page_width()
{
    PrintInfo * pi = [NXApp printInfo];			// Page layout details
    NXCoord topMargin, bottomMargin, leftMargin, rightMargin;
    const NXRect * paper = [pi paperRect];		//	In points
    
    [pi getMarginLeft:&leftMargin right:&rightMargin
    			top:&topMargin bottom:&bottomMargin];	/* In points */
    return (paper->size.width - leftMargin - rightMargin);

}

/*	Update the document and window according to the anchor info
*/

- syncWithAnchor
{
    CONST char * title = HTAnchor_title(nodeAnchor);
    if (title) [window setTitle:title];
    else {
        char * addr = HTAnchor_address((HTAnchor*)nodeAnchor);
	[window setTitle:addr];
	free(addr);
    }
    
    [self setEditable: containsUnimplemented ? NO 
    		: HTList_indexOf(HTAnchor_methods(nodeAnchor),
    				HTAtom_for("PUT")) != (-1) ? YES : NO];
    return self;
}

//			Class methods
//			-------------
//


//	Build a HyperText GIVEN its nodeAnchor.
//	--------------------------------------
//
//

- initAnchor:(HTParentAnchor *)anAnchor

{
    nextAnchorNumber = 0;
    protection = 0;			// Can do anything
    idBase = 0;
    StrAllocCopy(idBase, ANCHOR_ID_PREFIX);
    format = WWW_HTML;		// By default
    [self setMonoFont: NO];		// By default
    theRuns->chunk.growby = 16 * sizeof(theRuns->runs[0]); //  efficiency

    [self setDelegate: TheManager];			/* For changes */
    
    containsUnimplemented = NO;			// not yet
    
    nodeAnchor= anAnchor;
    HTAnchor_setDocument(anAnchor, (void*) self);
    [self syncWithAnchor];
    return self;
}

//	Set the next id to be used by the editor

- setNextId: (const char *) s
{
    const char * p = s;
    while (*p && (*p < '0' || *p > '9')) p++;  /* find numeric */
    if (!*p) return nil;	/* Syntax error: No number */
    nextAnchorNumber = atoi(p);
    
    if (p==s) return nil;	/* Syntax error: No alpha */
    if (idBase) free(idBase);
    idBase = (char *)malloc(p-s+1);
    strncpy(idBase, s, p-s);	/* Pick up prefix in case different */
    idBase[p-s] = 0;		/* and terminate it! */
    
    if (TRACE) fprintf(stderr, "HyperText: Next ID will be: %s%i\n",
    			idBase, nextAnchorNumber);
    return self;
}

 
 
+ newAnchor:(HTParentAnchor *)anAnchor
{
    NXRect aFrame = {{0.0, 0.0}, {page_width(), NICE_HEIGHT}};
    if (TRACE) printf("\nNew hypertext object\n");
    self = [super newFrame:&aFrame];
    [self initAnchor:anAnchor];
    return self;
}


//	Note: This one creates the new window if necessary too!
//
+ newAnchor:(HTParentAnchor *)anAnchor
		structure: 	(HTStructured*)s
		request:	(HTRequest*)request
{
    NXRect aFrame = {{0.0, 0.0}, {page_width(), NICE_HEIGHT}};
    if (TRACE) printf("\nNew hypertext object\n");
    if (request->context && HTMainText) {	/* @@@@ Reuse window */
        self = HTMainText;
	HTAnchor_setDocument(nodeAnchor, NULL);
        if (structure)
         (*structure->isa->free)(structure);	// release strcutured stream
	 
	[self initAnchor:anAnchor];
	adjustable = NO;		// Keep window size
    } else {
	self = [super newFrame:&aFrame];
	[self initAnchor:anAnchor];
	[self setupWindow];
	adjustable = YES;		// Can change window size.
    }
    [self setStructure:s];
    return self;
}


//	Free the hypertext.

- free
{
    slot[slotNumber] = 0;		//	Allow slot to be reused
    if (structure) 
      (*structure->isa->free)(structure);	// release structured stream
    HTAnchor_setDocument(nodeAnchor, NULL);
    if (idBase) free(idBase);
    if (HTMainText==self) {
        HTMainText = NULL;
	HTMainAnchor = NULL;
    }
    return [super free];
}

//	Reuse the hypertext for a different document  (used?)
//	--------------------------------------------

#ifdef UNUSED_CODE
- reuseAnchor:(HTParentAnchor *)anAnchor structure: (HTStructured*)s
{
    HTAnchor_setDocument(nodeAnchor, NULL);
    if (structure)
      (*structure->isa->free)(structure);	// release strcutured stream
 
    [self initAnchor:anAnchor];
    [self setStructure:s];
    return self;
}
#endif

- setContainsUnimplemented: (BOOL) whether  {
	containsUnimplemented = whether;
	return self;
}
- (BOOL) containsUnimplemented	{ return containsUnimplemented;}

//	Read and set format

- (HTFormat) format 			{return format; }
- setFormat:(HTFormat)f 		{format = f; return self; }

//	Pointer to structure input stream

- (HTStructured *) structure		{ return structure; }
- setStructure: (HTStructured *)s	{ structure = s; return self; }


//	Set the "info" pointer to the anchor instance for this run
//	----------------------

- setInfo: (NXRun*) r to: (void*)info
{
    if (r->rFlags.graphic) [(HyperCell*)(r->info) setLinkAnchor:info];
    else {
	r->info = info;
	r->rFlags.underline = NO;
	if (info && HTAnchor_followMainLink((HTAnchor *)info))
		r->rFlags.underline = YES;
	r->rFlags.dummy = (r->info != 0); /* Keep track for typingRun */
    }
    return self;
}

- (void *) info: (NXRun*) r;
{
    return (r->rFlags.graphic) ? (void*) [(HyperCell*)(r->info) linkAnchor]
    			       : r->info;
}


//	Useful diagnostic routine:  Dump to standard output
//	---------------------------------------------------
//
//	This first lists the runs up to and including the current run,
//	then it lists the attributes of the current run.
//
- dump:sender
{
    int pos;				/* Start of run being scanned */
    int sob=0;				/* Start of text block being scanned */
    NXRun * r = theRuns->runs;
    NXTextBlock * block = firstTextBlock;
    
    printf("Hypertext %p, selected(%i,%i)", (void*)self, sp0.cp, spN.cp);
    if (delegate) printf(", has delegate");
    printf(".\n");
    
    printf("    Frame is at (%f, %f, size is (%f, %f)\n",
    	frame.origin.x, frame.origin.y,
    	frame.size.width, frame.size.height);
    
    printf("    Text blocks and runs up to character %i:\n", sp0.cp);
    printf(" ch#,  chars, font-grey-RGB, para, infor, rFlags\n");
    for (pos = 0; pos<=sp0.cp; pos = pos+((r++)->chars)) {
	while(sob <= pos) {
	    printf(
	        "%5i: Block of %i/%lui characters at 0x%p starts `%10.10s'\n",
	    	sob, block->chars,  malloc_size(block->text), 
		(void*)block->text,
		(char*) block->text);
	    sob = sob + block->chars;
	    block = block->next;
	}
	printf( r->rFlags.graphic ? "G" : " ");
	printf( [self info:r] ? "A" : " ");
	printf( " ");
	printf("%5i: %3i of %p-%3.2f-%6x p=%p i=%p fl=%4x\n",
	    pos, r->chars,
	    (void*)r->font, r->textGray,(unsigned)(r->textRGBColor),
	     r->paraStyle,
	     (void *)(r->info),
	    *(short unsigned int*)&(r->rFlags));

    }
    r--;	/* Point to run for start of selection */

    printf("\n    Current run:\n\tFont name:\t%s\n", [r->font name]);
    {
        NXTextStyle *p = (NXTextStyle *)r->paraStyle;
	if (!p) {
	    printf("\tNo paragraph style!\n");
	} else {
	    int tab;
	    printf("\tParagraph style %p\n", p);
	    printf("\tIndents: first=%f, left=%f\n",
	    	p->indent1st, p->indent2nd);
	    printf("\tAlignment type=%i, %i tabs:\n",
	    	p->alignment, p->numTabs);
	    for (tab=0; tab<p->numTabs; tab++) {
	    	printf("\t    Tab kind=%i at %f\n",
			p->tabs[tab].kind, p->tabs[tab].x);
	    }
	}
    }
    printf("\n");
    return self;
}


//	Adjust Scrollers and Window size for current text size
//	------------------------------------------------------
//
//	The effect depends on whether the HyperTxet is glagged as adjustable.
//
//The scrollers are turned off if they possibly can be, to simplify the screen.
// If the text is editable, they have to be left on, although formatted text is
// allowed to wrap round, and so horizontal scroll bars are not necessary.
// The window size is adjusted as a function of the text size and scrollers.
//
//	@@ Bug: The resize bar should be removed if there are no scrollers.
//	This is difficult to do -- might have to make a new window.
//
- adjustWindow
{
#define MAX_WIDTH  paperWidth

    NXRect scroll_frame;
    NXRect old_scroll_frame;
    NXSize size;
    BOOL scroll_X, scroll_Y;			// Do we need scrollers?
    
    ScrollView* scrollview = [window contentView];// Pick up id of ScrollView
    float paperWidth = page_width();		// Get page layout width
    
    [self syncWithAnchor];
    
    if (adjustable) {	// (Keeping same window => don't resieze)
    
	[window disableFlushWindow];	// Prevent flashes
	
	[self setVertResizable:YES];	// Can change size automatically
	[self setHorizResizable:tFlags.monoFont];
	[self calcLine];			// Wrap text to current text size
	[self sizeToFit];			// Reduce size if possible.
	
	if (maxY > MAX_HEIGHT) {
	    scroll_Y = YES;
	    size.height = NICE_HEIGHT;
	} else {
	    scroll_Y = [self isEditable];
	    size.height = maxY < MIN_HEIGHT ? MIN_HEIGHT : maxY;
	}
    
	if (tFlags.monoFont) {
	    scroll_X = [self isEditable] || (maxX>MAX_WIDTH);
	    [self setNoWrap];
	} else {
	    scroll_X = NO;
	    [self setCharWrap:NO];				// Word wrap please 
	}

#ifdef RESIZE_X			// Don't like this -- too much jitter
	if (maxX > MAX_WIDTH) {
	    size.width = MAX_WIDTH;
	} else {
	    size.width = maxX < MIN_WIDTH ? MIN_WIDTH : maxX;
	}
#else
	size.width = MAX_WIDTH;
#endif

    // maxX is the length of the longest line.
    //	It only represnts the width of the page
    //	 needed if the line is quad left. If the longest line was
    //	centered or flush right, it may be truncated unless we resize
    //	it to fit.
    
	if (!scroll_X) {
	    [self sizeTo:size.width:maxY];
	    [self calcLine];
	    [self sizeToFit];		// Algorithm found by trial and error.
	}
    
    //	Set up the scroll view and window to match:
    
	[ScrollView getFrameSize:&scroll_frame.size
			    forContentSize: &size
			    horizScroller:	scroll_X
			    vertScroller:	scroll_Y
			    borderType: 	NX_LINE];
			    
	[scrollview setVertScrollerRequired:scroll_Y];
	[scrollview setHorizScrollerRequired:scroll_X];
    
    //	Has the frame size changed?
    
	[scrollview getFrame:&old_scroll_frame];
	if ( (old_scroll_frame.size.width != scroll_frame.size.width)
	    ||(old_scroll_frame.size.height != scroll_frame.size.height)) {
    
				    
    // Now we want to leave the top left corner of the window unmoved:
    
#ifdef OLD_METHOD	
	    NXRect oldframe;
	    [window getFrame:&oldframe];
	    [window sizeWindow:scroll_frame.size.width:scroll_frame.size.height];
	    [window moveTopLeftTo: oldframe.origin.x
				    : oldframe.origin.y + oldframe.size.height];
    #else
	    NXRect newFrame;
	    scroll_frame.origin.x = 150 + (slotNumber % 10)   * 30
					+ ((slotNumber/10)%3)* 40;
	    scroll_frame.origin.y = 185 + NICE_HEIGHT - scroll_frame.size.height
					- (slotNumber % 10)   * 20
					- ((slotNumber/10)%3)*  3;
	    [Window getFrameRect:&newFrame
		    forContentRect:&scroll_frame
		    style:NX_TITLEDSTYLE];	// Doesn't allow space for resize bar
	    newFrame.origin.y = newFrame.origin.y - 9.0;
	    newFrame.size.height = newFrame.size.height + 9.0; // For resize bar
	    [window placeWindow:&newFrame];	
#endif
	}
	
#ifdef VERSION_1_STRANGENESS
    //	In version 2, the format of the last run is overwritten with the format
    //	of the preceding run!
	{
	    NXRect frm;		/* Try this to get over "text strangeness" */
	    [self getFrame:&frm];      
	    [self renewRuns:NULL text:NULL frame:&frm tag:0];
	}
#endif
	[window reenableFlushWindow];
	
    } /* end if adjustable */
    
    [self calcLine];	/* Prevent messy screen -- needed if no resize? */
    [window display];	/* Ought to clean it up */

    return self;
    
} /* adjustWindow */


//	Set up a window in the current application for this hypertext
//	-------------------------------------------------------------

- setupWindow
{
    NXRect scroll_frame;				// Calculated later
    NXSize min_size = {300.0,200.0};			// Minimum size of text
    NXSize max_size = {1.0e30,1.0e30};			// Maximum size of text
   
    ScrollView * scrollview;
    NXSize nice_size = { 0.0, NICE_HEIGHT };		// Guess height
        
    nice_size.width = page_width();
    [ScrollView getFrameSize:&scroll_frame.size
    			forContentSize: &nice_size
			horizScroller:NO
			vertScroller:YES
			borderType: NX_LINE];

    {
    	int i;	/* Slot address */
	for(i=0; (i<SLOTS) && slot[i]; i++) ;	/* Find spare slot */
	if (i=SLOTS) i = (window_sequence = (window_sequence+1) % SLOTS);
	slot[i] = self;
	slotNumber = i;
	scroll_frame.origin.x = 150 + (slotNumber % 10)   * 30
				    + ((slotNumber/10)%3)* 40;
	scroll_frame.origin.y = 185 - (slotNumber % 10)   * 20
				    - ((slotNumber/10)%3)*  3;
     }
     
 //	Build a window around the text in order to display it.

#define NX_ALLBUTTONS 7  // Fudge -- the followin methos is obsolete in 3.0:    
    window = [Window new];
    [window			initContent:	&scroll_frame
    				style:		NX_RESIZEBARSTYLE
				backing: 	NX_BUFFERED
				buttonMask:	NX_ALLBUTTONS
				defer:		NO];		// display now				
    [window setDelegate:self];			// Get closure warning
    [window makeKeyAndOrderFront:self];		// Make it visible
    [window setBackgroundGray: 1.0];		// White seems to be necessary.
    
    scrollview = [ScrollView newFrame:&scroll_frame];
    [scrollview setVertScrollerRequired:YES];
    [scrollview setHorizScrollerRequired:NO];		// Guess.
    [[window setContentView:scrollview] free]; 	// Free old view, size new one.

						
    [scrollview setDocView:self];
    [self setOpaque:YES];			// Suggested in the book
    [self setVertResizable:YES];		// Changes size automatically
    [self setHorizResizable:NO];
    [self setMinSize:&min_size];		// Stop it shrinking to nought
    [self setMaxSize:&max_size];		// Stop it being chopped when editing
    [self notifyAncestorWhenFrameChanged: YES]; // Tell scrollview See QA 555
    [window display];				// Maybe we will see it now
    return self;
}


//		Return Instance Variables
//		-------------------------

- (HTParentAnchor*) nodeAnchor	{ return nodeAnchor; }
-  setNodeAnchor: (HTParentAnchor*) anc
{
    nodeAnchor = anc;
    return self;
}

/*	Return reference to a part of, or all of, this node
**	---------------------------------------------------
*/

//	Generate an anchor for a given part of this node, giving it an
//	arbitrary (numeric) name.

- (char *) new_unique_name
{
    static    char s[20];

    sprintf(s,"%s%i",ANCHOR_ID_PREFIX, nextAnchorNumber++);
    [delegate textDidChange:self];
    return s;
}


//	Check whether an anchor has been selected
//	-----------------------------------------

- (HTChildAnchor *) anchorSelected
{
    int sor;
    NXRun *r, *s, *e;		/* Scan, Start and end runs */
    HTChildAnchor * a;
    
    
    for (sor = 0, s=theRuns->runs;
    	sor+s->chars<=sp0.cp;
    	sor = sor+((s++)->chars)) ;
    for (e=s; sor+e->chars<spN.cp; sor = sor+ (e++)->chars);
    for(r=s; r<=e; r++) {
	if (a= [self info:r]) return a;
    }
    if (TRACE) printf("HyperText: No anchor selected.\n");
    return NULL;
}


//	Select an anchor
//	----------------
//
//	If there are any runs linked to this anchor, we select them. Otherwise,
//	we just bring the window to the front.

- selectAnchor:(HTChildAnchor *)anchor
{
    NXRun *s, *e;	/* run for start, run for end */
    int start, sor;
    NXRun * limit = (NXRun *)((char *)theRuns->runs + theRuns->chunk.used);
    for (sor=0, s=theRuns->runs; s<limit; sor = sor+(s++)->chars) {
        if (s->info == (void *)anchor){
		start = sor;
		
	        for (e=s; (e < limit)
				&&((e+1)->info == (void*)anchor);
			sor = sor+(e++)->chars);
    		[window makeKeyAndOrderFront: self];
		[self setSel:start:sor+e->chars];
		return [self scrollSelToVisible];
	    }
    }
    if (TRACE) printf("HT: Anchor has no explicitly related text.\n");
    return [window makeKeyAndOrderFront: self];
}


//

//	Return selected link (if any)				selectedLink:
//	-----------------------------
//
//

- (HTChildAnchor*) selectedLink
{

    int sor;			/* Start of run */
    NXRun *r, *s, *e;		/* Scan, Start and end runs */
    HTChildAnchor * a;
    int startPos, endPos;
    
    for (sor = 0, s=theRuns->runs;
    	sor+s->chars<=sp0.cp;
    	sor = sor+((s++)->chars)) ;
    startPos = sor;			/* Start of s */
    for (e=s; sor+e->chars<spN.cp; sor = sor+ (e++)->chars);
    for(r=s, a=NULL; r<=e; startPos = startPos + (r++)->chars) {
	a = [self info:r];
	if (a) break;
    }

    if (!a) {
        if (TRACE) printf("HyperText: No anchor selected.\n");
	return NULL;
    }
    
//	Extend/reduce selection to entire anchor

    {
	NXRun * limit = (NXRun *)((char *)theRuns->runs + theRuns->chunk.used);
	endPos = startPos + r->chars;
 	for (s=r; s > theRuns->runs && (HTChildAnchor *)[self info:(s-1)] == a;
			 s--)
		startPos = startPos-(s-1)->chars;
	for (e=r; e+1 < limit &&  (HTChildAnchor *)[self info:(e+1)] == a; e++)
		endPos   =   endPos+(e+1)->chars;
	[self setSel:startPos:endPos];
    }
    return a;
}


//	Similar but return Image anchor

- (HTChildAnchor*) selectedImageAnchor
{

    int sor;			/* Start of run */
    NXRun *r, *s, *e;		/* Scan, Start and end runs */
    HTChildAnchor * a;
    int startPos;
    
    for (sor = 0, s=theRuns->runs;
    	sor+s->chars<=sp0.cp;
    	sor = sor+((s++)->chars)) ;
    startPos = sor;			/* Start of s */
    for (e=s; sor+e->chars<spN.cp; sor = sor+ (e++)->chars);
    for(r=s, a=NULL; r<=e; startPos = startPos + (r++)->chars) {
	if (r->rFlags.graphic) {
	    id c = (HyperCell*)r->info;
	    if ([c class] == [ImageCell class]) { 
		a = (HTChildAnchor*)[c imageAnchor];
		if (a) break;
	    }
	}
    }

    if (!a) {
        if (TRACE) printf("HyperText: No image selected.\n");
	return NULL;
    }
    
//	Reduce selection to image only

    {
	[self setSel:startPos:startPos+r->chars];
    }

    return a;
}

///	Follow selected link (if any)				followLink:
//	-----------------------------
//

//	Find selected link and follow it

- (HTChildAnchor*)followLinkReuse: (BOOL)reuse
{
    HTChildAnchor * a = [self selectedLink];
    HTRequest * request = [TheManager request];
    request->context = (void*)(int)reuse;	/* better format */
    
    if (!a) return a;		// No link selected
   
    HTRequest_clear(request); // @@ Give Referer
    if(HTLoadAnchor(HTAnchor_followMainLink((HTAnchor*)a), request)) {
    	HTHistory_record((HTAnchor*)a);
	return a;
    }
    if (TRACE) printf("HyperText: Can't follow anchor.\n");
    return a;			// ... but we did highlight it.
}

//	Find selected link and follow it

- (HTChildAnchor*)followLink
{
    return [self followLinkReuse:NO];
}

- setTitle:(const char *)aString
{
    return [window setTitle:aString];
}


//				STYLE METHODS
//				=============
//

//	Find Unstyled Text
//	------------------
//
// We have to check whether the paragraph style for each run is one
// on the style sheet.
//
- selectUnstyled: (HTStyleSheet *)sheet
{
    NXRun * r = theRuns->runs;
    int sor;
    for (sor=0; sor<textLength; r++) {
        if (!HTStyleForParagraph(sheet, r->paraStyle)) {
	    [self setSel:sor:sor+r->chars];	/* Select unstyled run */
	    return self;
	}
        sor = sor+r->chars;
    }
    return nil;
}


//	Check styling
//	Is the styling all ok and sgml compatible?
//
- stylesOk
{
    NXRun * r = theRuns->runs;
    int sor;
    for (sor=0; sor<textLength; r++) {
        if (!HTStyleForParagraph(styleSheet, r->paraStyle)) {
	    return nil;
	}
        sor = sor+r->chars;
    }
    return self;
}


//	Copy a style into a run
//	-----------------------
- (void) applyStyle: (HTStyle *) style toRun: (NXRun *) r
{
    void * info;
    
    if (style->font) {
	r->font = style->font;
    }
    if (style->paragraph.lineHt) {
	r->paraStyle = &style->paragraph;
    }
    if (style->textGray>=0)
    	r->textGray = style->textGray;
	
    r->rFlags.underline = style->underline;

    if (style->anchor) {
	info = (style->anchor == CLEAR_POINTER) ? 0 : (style->anchor);
	[self setInfo:r to:info];
    } else {
        info = [self info:r];
    }
    
    
//    if (style->textRGBColor>=0)
    	r->textRGBColor = style->textRGBColor;
}


//	Apply an operation to a run
//	---------------------------

- (void) applyOperation: (HTOperation *) op toRun: (NXRun *) r
{
    void * info;
    HTNesting * old = ((HTStyle *) r->paraStyle)->nesting;
    HTNesting * nesting;
    HTStyle * style;
//    int old_level = HTMLLevel(old);
    nesting = op->insert ?
    		HTInsertLevel(old, op->element_number, op->level)
		: HTDeleteLevel(old, op->level);
    style = HTStyleForNesting(nesting);
    
/*	Apply style
*/
    if (style->font) {
	r->font = style->font;
    }
    if (style->paragraph.lineHt) {
	r->paraStyle = &style->paragraph;
    }
    info = op->info;
    if (op->set_info) {
	[self setInfo:r to:op->info];
    } else {
        info = [self info:r];
    }
    
    if (style->textGray>=0)
    	r->textGray = style->textGray;
	
    r->textRGBColor = style->textRGBColor;
    

}	

//	Check whether copying a style into a run will change it
//	-------------------------------------------------------

PRIVATE BOOL willChange(HTStyle * style, NXRun *r)
{
    if (r->rFlags.graphic) return YES;

    if (r->font != style->font) return YES;

//    if (style->textRGBColor>=0)
    	if (r->textRGBColor != style->textRGBColor) return YES;

    if (style->textGray>=0)
    	if (r->textGray != style->textGray) return YES;

    if (style->paragraph.lineHt) {
        if (r->paraStyle != &style->paragraph) return YES;
    }
    if (style->anchor) {
        if (r->info != style->anchor) return YES;
    }
    return NO;
}

//	Update a style
//	--------------
//
//
- updateStyle:(HTStyle *)style
{
    NXRun * r = theRuns->runs;
    int sor;
    for (sor=0; sor<textLength; r++) {
        if (r->paraStyle == &style->paragraph) 
	    [self applyStyle:style toRun:r];
        sor = sor+r->chars;
    }
    [self calcLine];
    [window display];
    return nil;
}

//	Delete an anchor from this node, without freeing it.
//	----------------
//
//
#ifdef OLD_CODE
- disconnectAnchor:(HTChildAnchor *)anchor
{
    NXRun * r = theRuns->runs;
    int sor;
    for (sor=0; sor<textLength; r++) {
        if ([self info:r] == (void *)anchor) 
	    [self setInfo:r to: 0];
	    r->textGray =  NX_BLACK;
        sor = sor+r->chars;
    }
    [window display];
    return nil;
}
#endif

//	Find start of paragraph
//	-----------------------
//
//	Returns the position of the character after the newline, or 0.
//
- (int) startOfParagraph: (int) pos
{
    NXTextBlock * block;
    int sob;
    unsigned char * p;
    for(block=firstTextBlock, sob=0; sob+block->chars <=pos; block=block->next)
        sob = sob + block->chars;
    for(p=block->text+(pos-sob)-1; p >= block->text; p--)
        if (*p == '\n') return sob + (p - block->text) +1 ;/* Position of newline */
    while (block->prior) {
        block = block->prior;
	sob = sob - block->chars;
        for(p=block->text+(block->chars-1); p>=block->text; p--)
            if (*p == '\n') return sob + (p-block->text)+1;/* Position of newline */
    }
    return 0;
}


//	Find end of paragraph
//	-----------------------
//
//	Returns the position after the newline, or the length of the text.
//	Note that any number of trailing newline characters are included in
//	this paragraph .. basically because the text object does not support
//	the concept of space after or before paragraphs, so extra paragrpah
//	marks must be used.
//
- (int) endOfParagraph: (int) pos
{
    NXTextBlock * block;
    int sob;
    unsigned char * p;
    BOOL found_newline = NO;
    
    if (pos>=textLength) return textLength;
    
    for(block=firstTextBlock, sob=0; sob+block->chars <=pos; block=block->next)
        sob = sob + block->chars;	// Find text block for pos
    
    p = block->text+(pos-sob);		// Start part way through this one
    
    while (block) {
        for(; p < block->text+block->chars; p++) {
            if (found_newline) {
	    	if (*p != '\n')
		    return sob + (p-block->text); /* Position after newline */
	    } else {
		if (*p == '\n') {
		    found_newline = YES;
		}
	    }
	}
	sob = sob + block->chars;	/* Move to next block */
        block = block->next;
	if (block) p = block->text;
	
    }
    return textLength;
}


//	Do two runs imply the same format?
//	----------------------------------

BOOL run_match(NXRun* r1, NXRun *r2)
{
    return 	(r1->font == r2->font)
    	&&	(r1->paraStyle == r2->paraStyle)
    	&&	(r1->textGray == r2->textGray)
    	&&	(r1->textRGBColor == r2->textRGBColor)
	&&	(r1->superscript == r2->superscript)
	&&	(r1->subscript == r2->subscript)
	&&	(r1->info == r2->info)
	&&	(*(int *)&r1->rFlags == *(int*)&r2->rFlags);
}

//	Check Consecutive runs and merge if necessary
//	---------------------------------------------
//
//	If the runs match in EVERY way, they are combined into one, and
//	all the other runs are shuffled down.
//
- (BOOL) mergeRun: (NXRun *) run
{
    NXRun * r, *last;
    if (run_match(run, run+1) ) {
        if (TRACE) printf("HT: Merging run %p\n", (void*)run);
        run->chars = run->chars +(run+1)->chars;
	last = ((NXRun *)((char*)theRuns->runs + theRuns->chunk.used))-1;
	for(r=run+1; r<last; r++) r[0] = r[1];
        theRuns->chunk.used = theRuns->chunk.used - sizeof(*r);
    	return YES;
    }
    return NO;
}


//	Common style for a given region
//	-------------------------------
//


- (HTNesting*) commonNestingFrom:(int)start to:(int)end
{
    int pos;				/* Character position within text */
    NXRun * s, *e;			/* Start and end run 		*/
    HTNesting * yca;			/* common style */
    
//	We determine in which runs the first and last characters to
//	be changed lie.
   
    for (pos=0, s=theRuns->runs; pos+s->chars<=start;
    			pos = pos+((s++)->chars)) /*loop*/;
/*	s points to run containing char after selection start */

    yca = ((HTStyle*)(s->paraStyle))->nesting;
    for (e=s; pos+e->chars < end; pos = pos+((e++)->chars)) {
        yca= HTMLCommonality(yca, ((HTStyle*)(e->paraStyle))->nesting);
    }

    return yca;
}


//	Apply OPERATION to a given region
//	---------------------------------
//
// 	Note that one should not have two consecutive runs of the same style,
// 	nor any zero length runs. We have a little calculation, therefore,
// 	in order to work out how many runs will eventually be needed:
//	this may be more or less than we started with.
//	Remember that appling a style to a run may or may not change it.
//
//	PS: Actually, we notice that text insertion does leave two consecutive
//	runs the same in the Text object, but deletion cleans up.


- applyOperation: (HTOperation*) op from:(int)start to:(int)end
{
    int pos;				/* Character position within text */
    int increase;			/* Number of runs to be split	*/
    int new_used;			/* New number of bytes in runs	*/
    BOOL need_run_before, need_run_after;/* Sometimes we don't need them	*/
    int	run_before_start, run_after_end;/* Start of run_before etc 	*/
    NXRun * s, *e;			/* Start and end run 		*/
    NXRun *p;				/* Pointer to run being read	*/
    NXRun *w;				/* Pointer to run being written	*/
    NXRun *r;				/* Pointer to end of runs	*/
    
    if (start == end) {
        [self applyOperation:op toRun:&typingRun];		/* Will this work? */
        if (TRACE) printf("Style applied to typing run\n");
	return nil;				/* Can't operate on nothing */
    }

//	First we determine in which runs the first and last characters to
//	be changed lie.
   
    for (pos=0, s=theRuns->runs; pos+s->chars<=start;
    			pos = pos+((s++)->chars)) /*loop*/;
/*	s points to run containing char after selection start */
    run_before_start = pos;

    for (e=s; pos+e->chars < end; pos = pos+((e++)->chars)) ;	/* Find end run */
/*	e points to run containing character before selection end */
    run_after_end = pos+e->chars;

    r = (NXRun *) (((char *)(theRuns->runs)) +theRuns->chunk.used);	/* The end*/
    
    if (TRACE) {
        printf("Runs: used=%i, elt. size=%i, %i elts, total=%li\n",
    		theRuns->chunk.used, sizeof(*r), r - theRuns->runs,
		(r-theRuns->runs)*sizeof(*r) );   
	printf("    runs at %p, r=%p. textLength:%i, r ends at:%i\n",
		theRuns->runs, r, textLength, pos);
    }
    
       
//	Move up runs as necessary in order to make room for the splitting
//	of the start and end runs into two.  We only do this if necessary.

//    if (!willChange(style, s))
//        start = run_before_start;	/* No run before is needed now */
    need_run_before = (start>run_before_start);
 
//    if (!willChange(style, e)) 
//       end = run_after_end;		/* No run after is needed now */
    need_run_after = (end < run_after_end);

    if (TRACE) printf(
    	"Run s=%i, starts at %i; changing (%i,%i); Run e=%i ends at %i\n",
    	s-theRuns->runs, run_before_start, start, end, e-theRuns->runs, run_after_end); 
    
    increase = need_run_after + need_run_before;
    if (increase) {
        new_used = theRuns->chunk.used + increase*sizeof(*r);   
	if (new_used> theRuns->chunk.allocated) {
	    NXRun* old = theRuns->runs;
	    theRuns = (NXRunArray*)NXChunkGrow(&theRuns->chunk, new_used);
	    if (theRuns->runs !=old) {			/* Move pointers */
		if (TRACE) printf("HT:Apply style: moving runs!\n");
		e = theRuns->runs + (e-old);
		r = theRuns->runs + (r-old);
		s = theRuns->runs + (s-old);
	    }
	}
	for (p=r-1; p>=e; p--) p[increase] = p[0];	/* Move up the runs after */
	r = r+increase;					/* Point to after them 910212*/
	/* p = e-1 */
	
	if (need_run_after) {
	    e = e + increase-1;				/* Point last to be changed */
	    e[0] = e[1];				/* Copy the last run */
	    e[1].chars = run_after_end - end;
	    e[0].chars = e[0].chars - e[1].chars;	/* Split the run into two */
	}
	
	if (need_run_before) {
	    for(;p>=s; p--) p[1] = p[0];		/* Move runs up, copying 1st*/
	    s[0].chars = start - run_before_start;	/* Split the run into two */
	    if (need_run_after && (s+1==e)) {		/* If only one middle run */
		s[1].chars = end-start;			/* The run we need */
	    } else {
	        s[1].chars = s[1].chars - s[0].chars;	/* The remainder */
	    }
	    s++;		/* Move on to point to first run to be changed */
	    if (!need_run_after) e++;			/* First to be changed */
	}
    	theRuns->chunk.used = new_used; 
	
    } /* end if increase */
    
// We consider the bit of text which is to be styled, s thru e.
// We scan through, first, applying the style, until we find two runs which
// need to be merged.

    p=s;
    if (p==theRuns->runs) {
        [self applyOperation:op toRun:p++];		/* Don't merge with run -1! */
    }
    
    while(p<=e) {
	[self applyOperation:op toRun:p++];
	if (run_match(p-1,p-2)) {
	    break;
        }
    }

//	Once we have merged two runs, we have to copy the rest of them across,
//	merging others as necessary.

    w = p-1;			/* w now points to last written run */
    for(;p<=e; p++) {
        [self applyOperation:op toRun:p];
	if (run_match(p,w)) {
	    w->chars = w->chars+p->chars;	/* Combine  with w */
	} else {
	    w++;				/* or skip */
	    *w = *p;				/* and keep a copy */
	}
    }

//	Now, is any runs were merged, we have to copy the rest of the runs down
//	and decrease the size of the chunk.

    w++;					/* Point to next to be written */
    if (w<p) {					/* If any were moved, */
	for(;p<r;) *w++ = *p++;			/* Move the following runs down */	
    	theRuns->chunk.used = (char*)w - (char*)theRuns->runs; 
    }
    
    [self calcLine];				/* Update line breaks */
    return [window display];			/* Update window */
}

//	Indent list level

- insertLevel: (int) element_number  info: (void*) info
{
    HTNesting * yca = [self commonNestingFrom:sp0.cp to:spN.cp];
    HTOperation operation;
    operation.insert = YES;
    operation.level = HTMLLevel(yca);
    operation.set_info = (element_number == HTML_A);
    operation.info = info;
    operation.element_number = element_number;
    [self applyOperation: &operation from:sp0.cp to:spN.cp];
    return self;
}

//	Delete list level

- deleteLevel: (int) element_number
{
    HTOperation operation;
    HTNesting * yca = [self commonNestingFrom:sp0.cp to:spN.cp];
    int i = HTMLLevel(yca);
    for (; i>0; i--) {
        if (HTMLElementNumber(yca) == element_number) break;
        yca = HTMLAncestor(yca, 1);
	if (!yca) return nil;		/* fail - no such element! */
    }
    operation.insert = NO;
    operation.set_info = (element_number == HTML_A);
    operation.info = NULL;
    operation.level = i;
    operation.element_number = element_number;
    [self applyOperation: &operation from:sp0.cp to:spN.cp];
    return self;
}

///////////////////////////// Anchor Specific

//	Public method:	Generate an anchor for the selected text
//	--------------------------------------------------------
//
//	If the document is not editable, then nil is returned
//	(unless the user asks for an existing anchor).
//
- (HTChildAnchor *) referenceSelected
{
    HTChildAnchor * a;

    
    a = [self anchorSelected];
    if (a) return a;			/* User asked for existing one */
    
    if ([self isEditable]) [window setDocEdited:YES];
    else return NULL;

    a = HTAnchor_findChildAndLink(nodeAnchor,
    	[self new_unique_name], NULL, NULL);
    [self insertLevel: HTML_A info:a];
    if (TRACE) printf("HyperText: New dest anchor %p from %i to %i.\n",
    	a, sp0.cp, spN.cp);
    [delegate textDidChange:self];
    return a;
}

//	Generate an anchor for the text, and link it to a given one
//	-----------------------------------------------------------

- (HTChildAnchor *) linkSelTo:(HTAnchor *)anAnchor
{
    HTChildAnchor * a;
    
    if (!anAnchor) return NULL;			/* Anchor must exist */
    
    if ([self isEditable]) [window setDocEdited:YES];
    else return NULL;

    a = [self anchorSelected];
    if (!a) {
	a = HTAnchor_findChildAndLink(nodeAnchor, [self new_unique_name],
		NULL, NULL);
    	if (TRACE) printf("HyperText: New source anchor %p from %i to %i.\n",
	 		a, sp0.cp, spN.cp);
    } else {
	[self selectAnchor:a];
    	if (TRACE) fprintf(stderr,"HyperText: Existing source anchor %p selected.\n",a);
    }
    HTAnchor_link((HTAnchor*)a, anAnchor, NULL); 	/* no type yet */
    [self insertLevel: HTML_A info:a];
    return a;
}


//	Purge anchor from selected text
//	-------------------------------
//
//	The anchor is left becuase in general we don't delete anchors.
//	In any case, we would have to check whether all text referencing it
//	was deleted.
//
- unlinkSelection
{
    if ([self isEditable]) [window setDocEdited:YES];
    else return nil;
    
    return [self deleteLevel: HTML_A];
}


- (HTParentAnchor *) referenceAll
{
    return nodeAnchor;	// Just return the same one each time   
}



//	Apply style to a given region
//	-----------------------------
//
// 	Note that one should not have two consecutive runs of the same style,
// 	nor any zero length runs. We have a little calculation, therefore,
// 	in order to work out how many runs will eventually be needed:
//	this may be more or less than we started with.
//	Remember that appling a style to a run may or may not change it.
//
//	PS: Actually, we notice that text insertion does leave two consecutive
//	runs the same in the Text object, but deletion cleans up.


- applyStyle:(HTStyle *)style from:(int)start to:(int)end
{
    int pos;				/* Character position within text */
    int increase;			/* Number of runs to be split	*/
    int new_used;			/* New number of bytes in runs	*/
    BOOL need_run_before, need_run_after;/* Sometimes we don't need them	*/
    int	run_before_start, run_after_end;/* Start of run_before etc 	*/
    NXRun * s, *e;			/* Start and end run 		*/
    NXRun *p;				/* Pointer to run being read	*/
    NXRun *w;				/* Pointer to run being written	*/
    NXRun *r;				/* Pointer to end of runs	*/
    
    if (start == end) {
        [self applyStyle:style toRun: &typingRun];	/* Will this work? */
        if (TRACE) printf("Style applied to typing run\n");
	return nil;				/* Can't operate on nothing */
    }

//	First we determine in which runs the first and last characters to
//	be changed lie.
   
    for (pos=0, s=theRuns->runs; pos+s->chars<=start;
    			pos = pos+((s++)->chars)) /*loop*/;
/*	s points to run containing char after selection start */
    run_before_start = pos;

    for (e=s; pos+e->chars < end; pos = pos+((e++)->chars)) ;	/* Find end run */
/*	e points to run containing character before selection end */
    run_after_end = pos+e->chars;

    r = (NXRun *) (((char *)(theRuns->runs)) +theRuns->chunk.used);	/* The end*/
    
    if (TRACE) {
        printf("Runs: used=%i, elt. size=%i, %i elts, total=%li\n",
    		theRuns->chunk.used, sizeof(*r), r - theRuns->runs,
		(r-theRuns->runs)*sizeof(*r) );   
	printf("    runs at %p, r=%p. textLength:%i, r ends at:%i\n",
		theRuns->runs, r, textLength, pos);
    }
    
       
//	Move up runs as necessary in order to make room for the splitting
//	of the start and end runs into two.  We only do this if necessary.

    if (!willChange(style, s))
        start = run_before_start;	/* No run before is needed now */
    need_run_before = (start>run_before_start);
 
    if (!willChange(style, e)) 
        end = run_after_end;			/* No run after is needed now */
    need_run_after = (end < run_after_end);

    if (TRACE) printf(
    	"Run s=%i, starts at %i; changing (%i,%i); Run e=%i ends at %i\n",
    	s-theRuns->runs, run_before_start, start, end, e-theRuns->runs, run_after_end); 
    
    increase = need_run_after + need_run_before;
    if (increase) {
        new_used = theRuns->chunk.used + increase*sizeof(*r);   
	if (new_used> theRuns->chunk.allocated) {
	    NXRun* old = theRuns->runs;
	    theRuns = (NXRunArray*)NXChunkGrow(&theRuns->chunk, new_used);
	    if (theRuns->runs !=old) {			/* Move pointers */
		if (TRACE) printf("HT:Apply style: moving runs!\n");
		e = theRuns->runs + (e-old);
		r = theRuns->runs + (r-old);
		s = theRuns->runs + (s-old);
	    }
	}
	for (p=r-1; p>=e; p--) p[increase] = p[0];	/* Move up the runs after */
	r = r+increase;					/* Point to after them 910212*/
	/* p = e-1 */
	
	if (need_run_after) {
	    e = e + increase-1;				/* Point last to be changed */
	    e[0] = e[1];				/* Copy the last run */
	    e[1].chars = run_after_end - end;
	    e[0].chars = e[0].chars - e[1].chars;	/* Split the run into two */
	}
	
	if (need_run_before) {
	    for(;p>=s; p--) p[1] = p[0];		/* Move runs up, copying 1st*/
	    s[0].chars = start - run_before_start;	/* Split the run into two */
	    if (need_run_after && (s+1==e)) {		/* If only one middle run */
		s[1].chars = end-start;			/* The run we need */
	    } else {
	        s[1].chars = s[1].chars - s[0].chars;	/* The remainder */
	    }
	    s++;		/* Move on to point to first run to be changed */
	    if (!need_run_after) e++;			/* First to be changed */
	}
    	theRuns->chunk.used = new_used; 
	
    } /* end if increase */
    
//	We consider the bit of text which is to be styled, s thru e.
//	We scan through, first, applying the style, until we find two runs which
//	need to be merged.

    p=s;
    if (p==theRuns->runs) {
        [self applyStyle:style toRun:p++];	/* Don't merge with run -1! */
    }
    
    for(; p<=e; p++) {
	[self applyStyle:style toRun:p];
	if (run_match(p,p-1)) {
	    break;
        }
    }

//	Once we have merged two runs, we have to copy the rest of them across,
//	merging others as necessary.

    w = p-1;			/* w now points to last written run */
    for(;p<=e; p++) {
        [self applyStyle:style toRun:p];
	if (run_match(p,w)) {
	    w->chars = w->chars+p->chars;	/* Combine  with w */
	} else {
	    w++;				/* or skip */
	    *w = *p;				/* and keep a copy */
	}
    }

//	Now, is any runs were merged, we have to copy the rest of the runs down
//	and decrease the size of the chunk.

    w++;					/* Point to next to be written */
    if (w<p) {					/* If any were moved, */
	for(;p<r;) *w++ = *p++;			/* Move the following runs down */	
    	theRuns->chunk.used = (char*)w - (char*)theRuns->runs; 
    }
    
    [self calcLine];				/* Update line breaks */
    return [window display];				/* Update window */
}


//	Apply a style to the selection
//	------------------------------
//
//	If the style is a paragraph style, the
//	selection is extended to encompass a number of paragraphs.
//
- applyStyle:(HTStyle *)style
{
    int start, end;
    if (TRACE) printf("Applying style %p to (%i,%i)\n",
    		style, sp0.cp, spN.cp);
    
    if (sp0.cp<0) {					/* No selection */
    	return [self applyStyle:style from:0 to:0];	/* Apply to typing run */
    }
    
    if (!style) return nil;

    if ([self isEditable]) [window setDocEdited:YES];
    else return nil;

    	
    start = sp0.cp;
    end = spN.cp;
    if (style->paragraph.lineHt) { /* Extend to an integral number of paras. */
        start = [self startOfParagraph:start];
        end = [self endOfParagraph:end];
    }
    return [self applyStyle:style from:start to:end];
}


//	Apply style to all similar text
//	-------------------------------
//
//
- applyToSimilar:(HTStyle*)style
{
    NXRun * r = theRuns->runs;
    int sor;
    NXRun old_run;
    
    for (sor = 0; sor<=sp0.cp; sor = sor+((r++)->chars)) ;/* Find run after */
    old_run = *(r-1);			/* Point to run for start of selection */

    if (TRACE) printf(
    	"Applying style %p to unstyled text similar to (%i,%i)\n",
    	style, sp0.cp, spN.cp);

    for(r=theRuns->runs;
    	 (char*)r-(char*)theRuns->runs < theRuns->chunk.used; r++) {
        if (r->paraStyle == old_run.paraStyle) {
	    if(TRACE) printf("    Applying to run %p\n", r);
	    [self applyStyle:style toRun:r];
	    if (r!=theRuns->runs){
	    	if ([self mergeRun:r-1]) r--;	/* Do again if shuffled down */
	    }
	}
    }
    [self calcLine];
    [window display];
    return self;
}

//	Pick up the style of the selection
//	----------------------------------

- (HTStyle *) selectionStyle:(HTStyleSheet *) sheet
{
    return HTStyleForNesting([self commonNestingFrom:sp0.cp to:spN.cp]);
    
#ifdef PREVIOUSLY
    NXRun * r = theRuns->runs;
    int sor;
    for (sor = 0; sor<=sp0.cp; sor = sor+((r++)->chars)) ;/* Find run after */
    r--;				/* Run for start of selection */
    return HTStyleForRun(sheet, r);	/* for start of selection */
#endif    
}


//	Another replaceSel method, this time using styles:
//	-------------------------------------------------
//
//	The style is as given, or where that is not defined, as the
//	current style of the selection.

- replaceSel: (const char *)aString style:(HTStyle*)aStyle
{
    NXRun * r = theRuns->runs;
    int sor;
    NXRunArray	newRuns;
    
    for (sor = 0; sor<=sp0.cp; sor = sor+((r++)->chars)) ;/* Find run after */
    r--;					/* Run for start of selection */
    newRuns.runs[0] = *r;			/* Copy it */
    newRuns.chunk.used = sizeof(*r);		/* 1 run used */
    [self applyStyle:aStyle toRun:newRuns.runs];	/* change it */
    newRuns.runs->chars = strlen(aString);	/* Match the size to the string */
    return [self replaceSel:aString length:newRuns.runs->chars runs:&newRuns];
}


//	Read in as Plain Text					readText:
//	---------------------
//
//	This method overrides the method of Text, so as to force a plain text
//	hypertext to be monofont and fixed width.  Also, the window is updated.
//
- readText: (NXStream *)stream
{
//    [self setMonoFont:YES];		Seems to leave it in a strange state
    [self setHorizResizable:YES];
    [self setNoWrap];
    [self setFont:[Font newFont:"Ohlfs" size:10.0]];	// @@ Should be XMP
    [super readText:stream];
    format = WWW_PLAINTEXT;				// Remember
    
#ifdef NOPE
    {
      NXRect frm;		/* Try this to get over "text strangeness" */
      [self getFrame:&frm];      /* on plain text only Aug 91 */
      [self renewRuns:NULL text:NULL frame:&frm tag:0];
    }
#endif
    [self adjustWindow];
    return self;
}


//	Read in as Rich Text					readRichText:
//	---------------------
//
//	This method overrides the method of Text, so as to force a plain text
//	hypertext to be monofont and fixed width.  Also, the window is updated.
//
- readRichText: (NXStream *)stream
{
    id status =  [super readRichText:stream];
    [self adjustWindow];
    format = WWW_RICHTEXT;				// Remember
    return status;
}

////////////////////////////////////////////////////////////////

//				Window Delegate Methods
//				=======================

//	Prevent closure of edited window without save
//
- windowWillClose:sender
{
    int choice;
    if (![window isDocEdited]) return self;
    choice = NXRunAlertPanel("Close", "Save changes to `%s'?",
    	"Yes", "No", "Don't close", [window title]);
    if (choice == NX_ALERTALTERNATE) return self;
    if (choice == NX_ALERTOTHER) return nil;

    return nil; //[server saveNode:self];    @@@@@@@@ server!!!
}

//	Change configuration as window becomes key window
//
- windowDidBecomeMain:sender
{
    return [delegate hyperTextDidBecomeMain:self];
}

/*		Writing to this object as a styled stream
**		=========================================
**
*/
/*		Input procedure for printing a trace as we go
*/
/*#define NEXT_CHAR NXGetc(sgmlStream)
  #define BACK_UP NXUngetc(sgmlStream)
*/

//	Inputting from the text object:
//	------------------------------



- (unsigned char) next_input_block
{
    char c = *read_pointer;
    read_block = read_block->next;
    if (!read_block) read_block = firstTextBlock;	/* @@@ FUDGE */
    read_pointer = read_block->text;
    read_limit = read_pointer + read_block->chars;
    return c;
}

///////////////////////////////////////////////////////////////////////


//			Outputting to the text object
//			=============================
//
//	These macros are used by the parse routines
//
#define BLOCK_SIZE NX_TEXTPER			/* Match what Text seems to use */



#define OUTPUT(c)	{ *write_pointer++ = (c); \
 if (write_pointer==write_limit) {[self end_output]; [self append_start_block]; }}
#define OUTPUTS(string)	{const char * p; for(p=(string);*p;p++) OUTPUT(*p);}

#define LOADPLAINTEXT	loadPlainText()


//	Allocate a text block to accumulate text
//
// Bugs:
// It might seem logical to set the "malloced" bit to 1, because the text block
// has been allocted with malloc(). However, this crashes the program as at the
// next edit of the text, the text object frees the block while still using it.
// Chaos results, sometimes corrupting the stack and/or looping for ages. @@
// We therefore set it to zero! (This might have been something else -TBL)
//
- append_start_block
{	
    NXTextBlock *previous_block=write_block;	/* to previous write block */
    
    if (TRACE)printf("    Starting to append new block.\n");
        
    lastRun = ((NXRun*) ((char*)theRuns->runs + theRuns->chunk.used))-1;
    write_block = (NXTextBlock*)malloc(sizeof(*write_block));
    write_block->tbFlags.malloced=0;		/* See comment above */
    write_block->text = (unsigned char *)malloc(BLOCK_SIZE);
    write_block->chars = 0;			// For completeness: not used.
    write_pointer = write_block->text;
    write_limit = write_pointer + BLOCK_SIZE;

//	Add the block into the linked list after previous block:

    write_block->prior = previous_block;
    write_block->next = previous_block->next;
    if (write_block->next) write_block->next->prior = write_block;
        else lastTextBlock = write_block;
    previous_block->next = write_block;
    return self;
}


//	Add a new run
//	-------------

- extendRuns
{
    int new_used = (((char *)(lastRun+2)) - (char*)theRuns->runs);
    if (new_used > theRuns->chunk.allocated) {
	if (TRACE) fprintf(stderr,
		" HT: Extending runs to %d bytes.\n", new_used); 
	theRuns = (NXRunArray*)NXChunkGrow(
		&theRuns->chunk, new_used);
	lastRun = ((NXRun*) ((char*)theRuns->runs +
				    theRuns->chunk.used))-1;
    }
    lastRun[1]=lastRun[0];
    lastRun++;
    theRuns->chunk.used = new_used;
    lastRun->chars = 0;			/* None yet */
    return self;
}

// 	Start append
//	------------
//
//		Does NOT clear object.
//
- append_continue
{
    if (TRACE)printf("Begin append to text.\n");
    
    original_length = textLength; 
    if (TRACE) fprintf(stderr,
      "Text originally contains %i characters\n", original_length);

        
    lastRun = ((NXRun*) ((char*)theRuns->runs +
    			theRuns->chunk.used))-1;

    if (lastRun->rFlags.graphic) {
    	fprintf(stderr, "HT: ERROR: How did it text end in a graphic?\n");
        [self extendRuns];
	lastRun->rFlags.subclassWantsRTF = 0;
	lastRun->rFlags.graphic = 0;
	lastRun->rFlags.forcedSymbol = 0;
	lastRun->rFlags.RESERVED = 0;
	[self applyStyle:&appendingStyle toRun:lastRun];
	[self setInfo: lastRun to:appendingInfo];
    }

    
//	Chop off the conventional trailing newline:

    {
    	int n;
	n = lastTextBlock->chars;
	if (n>0 && lastRun->chars>0 && lastTextBlock->text[n-1] == '\n') {
	    lastTextBlock->chars--;
	    lastRun->chars--;
	    textLength--;
	} else {
	    fprintf(stderr,"HT: ERROR: Can't chop trailing newline off!\n");
	}
    }

//	Use the last existing text block:


    write_block = lastTextBlock;
    write_pointer = write_block->text+write_block->chars;
    write_limit = write_block->text + BLOCK_SIZE;
    textLength = textLength - write_block->chars;  /* Added back on end */

    return self;
}

// 	Clear text object and start append
//	----------------------------------
//
- appendBegin
{

    if (TRACE)printf("Begin append to text.\n");
    
    [self setText:""];				// Delete everything there
    original_length = textLength; 
    if (TRACE) printf("Text now contains %i characters\n", original_length);

        
    lastRun = ((NXRun*) ((char*)theRuns->runs + theRuns->chunk.used))-1;

//	Use the last existing text block:

    write_block = lastTextBlock;
    
// It seems that the Text object doesn't like to be empty: it always wants to
// have a newline in at least. However, we need it seriously empty and so we
// forcibly empty it. CalcLine will crash if called with it in this state.

#ifdef OLD_CODE
    if (original_length==1) {
        if (TRACE) printf("HT: Clearing out single character from Text.\n");
        lastRun->chars =  0;		/* Empty the run */
	write_block->chars = 0;		/* Empty the text block */
	textLength = 0;		/* Empty the whole Text object */
	original_length = 0;		/* Note we have cleared it */
    }
#else  /* new */
    {
    	int n;
	n = lastTextBlock->chars;
	if (n>0 && lastRun->chars>0 && lastTextBlock->text[n-1] == '\n') {
	    lastTextBlock->chars--;
	    lastRun->chars--;
	    textLength--;
	} else {
	    fprintf(stderr,"HT: ERROR: Can't chop trailing newline off!\n");
	}
    }
#endif

    write_pointer = write_block->text+write_block->chars;
    write_limit = write_pointer + BLOCK_SIZE;
    return self;
}




//	Set a style for new text while appending
//	----------------------------------------

- appendStyle:(HTStyle *) style
{
    appendingStyle = *style;		/* take a copy */

    if (!style) {
        if (TRACE) printf("appendStyle: style is null!\n");
	return NULL;
    }
    if (TRACE) printf("    Changing to style `%s' -- %s change.\n",
    	style->name, willChange(style, lastRun) ? "will" : "won't");
    if (willChange(style, lastRun)) {
        int size;
	/*  paragraph_gap(style); */
	size = (write_pointer - write_block->text);
	lastRun->chars = lastRun->chars + size - write_block->chars;
	write_block->chars = size;
        if (lastRun->chars) [self extendRuns];
	[self applyStyle:style toRun:lastRun];
    }

    return self;
}


//	Set a info for new text while appending
//	----------------------------------------
- setAppendingInfo:(HTChildAnchor *)info
{

    int size;
    if (TRACE) printf("    Changing info to %p\n", info);

    size = (write_pointer - write_block->text);
    lastRun->chars = lastRun->chars + size - write_block->chars;
    write_block->chars = size;
    if (lastRun->chars) [self extendRuns];


    [self setInfo:lastRun to:info];

    appendingInfo = info;
    return self;
}


//	Transfer text to date to the Text object
//	----------------------------------------
- end_output
{
    int size = (write_pointer - write_block->text);
    if (TRACE)printf(
    	"    HT: Adding block of %i characters, starts: `%.20s...'\n",
    	size, (char*)write_block->text);
    lastRun->chars = lastRun->chars + size - write_block->chars;
    write_block->chars = size;
    textLength = textLength + size;
    return self;
}



//	Methods enabling an external parser to add styled data
//	------------------------------------------------------
//
//	These use the macros above in the same way as a built-in parser would.
//

- appendText: (const char *)text
{
    OUTPUTS(text);
    return self;
}

/*	Append character takes care of white space for paragraph marks.
*/
- appendCharacter: (char)ch
{
    if ((ch=='\n') && appendingStyle.paragraph.lineHt) {

	int newlines = (
	  ((appendingStyle.spaceBefore > appendingStyle.spaceAfter) ?
	  	 appendingStyle.spaceBefore : appendingStyle.spaceAfter) /
			    appendingStyle.paragraph.lineHt)         + 1;
	int i;
	for(i=0; i<newlines; i++) OUTPUT('\n');	/* approximate! @@	*/
 
    } else OUTPUT(ch);
    return self;
}


/*	Finish off and flush write to object
*/
- appendEnd
{
    int size;
//	Do this anyway. Always put one one, always take one off at start.
//    if (write_pointer == write_block->text && !write_block->prior)
        OUTPUT('\n');	/* Avoid empty object */

    size = write_pointer - write_block->text;
    if (size==0) {
        if (write_block->prior) {	/* If there are other blocks */
	    lastTextBlock = write_block->prior;/* Remove empty text block */
	    write_block->prior->next = 0;
	    free(write_block->text);
	    free(write_block);
	}
    } else {
	[self end_output];
    }

// get rid of zero length run if any

    if (lastRun->chars==0) {		/* Chop off last run */
        theRuns->chunk.used= (char*)lastRun - (char*)theRuns->runs;
    }
    
//	calcLine requires that the last character be a newline!
//#ifdef KLUDGE21
    {
        unsigned char * p = lastTextBlock->text +
			    lastTextBlock->chars - 1;
	if (*p != '\n') {
	    if (TRACE)
	    printf(
	    "HT: ERROR: Last character was %i not newline: overwriting!\n",
	    	*p);
	    *p = '\n';
	}
    }
//#endif
    [self adjustWindow];	/* Adjustscrollers and window size */

    return self;
}

//	Begin an anchor

- (HTChildAnchor *)appendBeginAnchor: (HTChildAnchor *) a
{
    HTStyle * style = HTStyleNew();

    style->anchor = a;
    HTAnchor_makeLastChild((HTChildAnchor *)style->anchor); /* Put in correct order */	      
    [self setAppendingInfo:a];		/* Start anchor here */
    free(style);
    return a;
}


- appendEndAnchor			// End it
{
    [self setAppendingInfo: NULL];		/* End anchor here */
    return self;
}

- appendObject: (int)	element_number		// Insert something non-text
	present: (BOOL*) present
	value:   (CONST char * CONST *) value
{

    HyperCell * cell;
 
    [self appendEnd];

    cell = [HyperCell new];
    [cell initForElement: element_number
	    inDTD: &HTMLP_dtd
	    attributesPresent:present
	    withValues:value
	    inView:self];

    [self setSel:textLength-1:textLength-1];	/* Insert before dummy NL */
    [self replaceSelWithCell:cell];
    [self append_continue];			/* contine appending */

    [self applyStyle:&appendingStyle toRun:lastRun-1];	/* Fix graphic run  */
    [self applyStyle:&appendingStyle toRun:lastRun];
    [self setInfo: lastRun-1 to:appendingInfo];
    [self setInfo: lastRun to:appendingInfo];
        
    return self;
    

}

- appendImage: (HTChildAnchor *) anc
	alternative: (CONST char *) alt
	aligned: (CONST char *) alignment
	isMap:   (BOOL)		ismap
{
    ImageCell * cell;
    
    if (!HTSuppressImages) {


#define USE_REPLACE_SEL		/* try it off one day (fails) */
#ifdef USE_REPLACE_SEL
	[self appendEnd];
    
	cell = [ImageCell new];
	[cell initAnchor:anc alternative:alt aligned:alignment isMap:ismap
	    inView:self];
	
    //    [cell setLinkAnchor:(HTChildAnchor*)old_run.info];
	
	[self setSel:textLength-1:textLength-1];/* Insert before dummy NL */
	[self replaceSelWithCell:cell];
	[self append_continue];			/* contine appending */
    
	[self applyStyle:&appendingStyle toRun:lastRun-1];/* Fix graphic  */
	[self applyStyle:&appendingStyle toRun:lastRun];
	[self setInfo: lastRun-1 to:appendingInfo];
	[self setInfo: lastRun to:appendingInfo];
#else
        [self extendRuns];
	[self appendCharacter: '\001'];	/// @@@ correct?  stand-in for graphic
	[self extendRuns];

	cell = [ImageCell new];
	[cell initAnchor:anc alternative:alt aligned:alignment isMap:ismap
	    inView:self];
	lastRun[-1].rFlags.graphic = YES;
	lastRun[-1].info = (void*)cell;		/// Cell
	[self setInfo: lastRun-1 to:appendingInfo];	//  Link stored in cell
#endif        
    
    } else {	/* suppressed image */
    
	[self appendBeginAnchor:anc];			/* @@@ placeholder */
	[self appendText:alt ? alt : ismap ? "CLICKABLE-IMAGE" : "IMAGE"];
	[self appendEndAnchor];
    }
    return self;

}


- replaceSelWithImage: (HTChildAnchor *) anc
	alternative: (CONST char *) alt
	aligned: (CONST char *) alignment
	isMap:   (BOOL)		ismap
{
    ImageCell * cell;
    
    cell = [ImageCell new];
    [cell initAnchor:anc alternative:alt aligned:alignment isMap:ismap
	inView:self];
    
    [self prePaste];
    [self replaceSelWithCell:cell];
    [self postPaste];
        
    return self;

}


//			Regenerating HTML
//			=================


#define PUTS(s) (*targetClass.put_string)(target, s)
#define START(e) (*targetClass.start_element)(target, e, 0, 0)
#define END(e) (*targetClass.end_element)(target, e)


struct _HTStream {
	CONST HTStreamClass *	isa;
	/* ... */
};


/*		Write SGML File back OUT
**		------------------------
**
**	This is currently quite NeXT-specific.
**
**	We run through te runs. When a characteristic of a run changes, we
**	output the approporiate SGML code. 
**
**	Bugs:	@@@ Highlighting is ignored.
**		@@@ end text is ignored.
*/


/*	Change Run
**	==========
*/
/*	This function generates the code for one run, given the previous run.
**
*/
- (void) changeRunFrom: (NXRun *) last to: (NXRun *) r relativeTo:(const char *)aName
{

   int chars_left = r->chars;
    
    HTStyle* old_style = (HTStyle*) last->paraStyle;
    HTStyle* new_style = (HTStyle*) r->paraStyle;
    
    HTNestingChange(
    	target,
    	old_style ? old_style->nesting : NULL,
	new_style ? new_style->nesting : NULL,
    	(HTChildAnchor *)[self info:r],
	aName);


/*	Now output the body of the run
*/

    if (r->rFlags.graphic) {
        [(HyperCell*)(r->info)  writeSGML:target relativeTo:aName];
	read_pointer++;	/* skip image's place holding character */
    } else {
	while (chars_left) {	
	    char c = (read_pointer+1==read_limit? 	[self next_input_block]
						:	*read_pointer++);
	    chars_left--;
	    HTRegenCharacter(c, new_style->nesting, target);
	}
    }
    
} /* changeRunFrom:to: */



/*	This is the body of the SGML output method.
*/
- writeSGML:(HTStructured *) thisTarget relativeTo:(const char *)aName
{
    NXRun * r = theRuns->runs;
    int sor;				/* Character position of start of run */
    NXRun dummy;    

    dummy.paraStyle = 0;
    dummy.info = 0;
    dummy.chars = 0;
    
    if (![self stylesOk]) {
	int choice = NXRunAlertPanel(
		"Contains unstyled text!",
		"Use \"Find unstyled text\" in the Style panel to find it."
		" Do you want to continue and possibly write strange HTML?",
	    "Yes", "No", NULL, [window title]);
	if (choice == NX_ALERTALTERNATE) return nil;
	if (choice == NX_ALERTOTHER) return nil;
    }
    
    target = thisTarget;	/* save in instance variable */
    targetClass = *target->isa; /* copy access routines */

    HTRegenInit();		/* Initialize regeneration of HTML */    

    
    read_block = firstTextBlock;
    read_pointer = read_block->text;	/* next character to be read */
    read_limit = read_pointer+read_block->chars;
    		
    START(HTML_HTML);
    START(HTML_HEAD);
    START(HTML_TITLE);
    PUTS([window title]);
    END(HTML_TITLE);
    
    if (nextAnchorNumber) {
	char * string = malloc(strlen(idBase + 10));
	sprintf(string, "%s%i", idBase, nextAnchorNumber);
	HTNextID(target, string);
	free(string);
    }
    
    END(HTML_HEAD);

/*	Change style tags etc
*/
    [self changeRunFrom:&dummy to:r relativeTo:aName];	/* Start first run */
    
    for (sor=r++->chars; sor<textLength; sor=sor+(r++)->chars)  {
        if (TRACE) printf("%4i:  %i chars in run %3li.\n",
			sor, r->chars, r-theRuns->runs);
	[self changeRunFrom:r-1 to: r relativeTo:aName];/* Runs 2 to N */
    }
    [self changeRunFrom:--r to:&dummy relativeTo:aName]; /* Close last run */

    tFlags.changeState = 0; 		/* Please notify delegate if changed */

    END(HTML_HTML);
    target->isa->free(target);
    
    return  self;    
}



//			Methods overriding Text methods
//			===============================
//
//	Respond to mouse events
//	-----------------------
//
// The first click will have set the selection point.  On the second click,
// we follow a link if possible, otherwise we allow Text to select a word
// as usual.
//
// We ruse the same window unless the logical ALTERNATE key is held down
// for double click (in fact the second of the two I guess).
//
- mouseDown:(NXEvent*)theEvent
{
    BOOL reuse = ((theEvent->flags & NX_ALTERNATEMASK) != 0) == HTMultiWindowMode;
 
    if (theEvent->data.mouse.click != 2) return [super mouseDown:theEvent];
    if (![self followLinkReuse: reuse  & ![window isDocEdited]])
    	return [super mouseDown:theEvent];
    return self;
}


//	The following are necessary to undo damage done by the Text object
//	in version 2.0 of NeXTStep. For some reason, iff the "info" is
//	nonzero, text typed in is given
//	a different copy of the typingRun parastyle, and zero anchor info.
//	Otherwise, the current run is broken in two at the insertion point,
//	but no changes made to the run contents.
//	The problem with simply repairing is that many runs will be made inside
//	an anchor.
//	We have to use a "dummy" flag to mean "This has an anchor: be careful!"
//	This is horrible.

- keyDown:(NXEvent*)theEvent
#ifdef TRY1
{
    id result;
    NXTextStyle *typingPara = typingRun.paraStyle;
    
    [self prePaste];
    
    result = [super keyDown:theEvent];
    
    {
	int inserted = originalEnd-originalStart +
			textLength-originalLength;
	
	if (TRACE) printf(
    "KeyDown, size(sel) %i (%i-%i)before, %i (%i-%i)after.\n",
	    originalLength, originalStart, originalEnd,
	    textLength, sp0.cp, spN.cp);
	    
	if (inserted>0) {
	    NXRun * s;
	    int pos;
	    int start = sp0.cp-inserted;
	    for (pos=0, s=theRuns->runs; pos+s->chars<=start;
		    pos = pos+((s++)->chars)) /*loop*/;

//	s points to run containing first char of insertion

	    if (pos!=start)
		    printf(
	    "HT: Strange: inserted %i at %i, start of run=%i !!\n",
				    inserted, start, pos);
				    
	    if (s > theRuns->runs) {	/* ie s-1 is valid */
		s->paraStyle = typingPara;	/* Repair damage to runs */
		/* What about freeing the old paragraph style? @@ */
		s->info = (s-1)->info;
		s->rFlags.dummy = 1;	/* Pass on flag */
	    }
	    
	}
    }
    return result;
}
#else
//	The typingRun field does not seem to reliably reflect the
//	format which would be appropriate if typing were to occur.
//	We have to use our own.
{
    NXRun run;
    void * info;
    {
	NXRun * s;	/* To point to run BEFORE selection */
	int pos;
	BOOL take_style_of_following_character = YES;
	
/* 	If there is a nonzero selection, take the run containing the
**	first character. If the selection is empty, take the run containing the
**	character before the selection, unless that is non-printing and
**	the character after it is printing, in which case take the style
**	of the character after.
*/
	if (sp0.cp == spN.cp) {
	    int q = 0;
	    char before, after;
	    NXTextBlock * block = firstTextBlock;
	    while(q+block->chars < sp0.cp) {
		q=q+block->chars;
		block = block->next;
	    }
	    before = block->text[sp0.cp - q - 1];
	    if (q+block->chars == sp0.cp) {	/* exactly on the edge */
		q=q+block->chars; 
		block = block->next;
	    }
	    after = block->text[sp0.cp - q];
	    take_style_of_following_character =
			    ((before <= ' ')  && (after > ' '));
	}
	
	if (!take_style_of_following_character) {
	    for (pos=0, s=theRuns->runs; pos+s->chars<sp0.cp;	/* Before */
		pos = pos+((s++)->chars)) /*loop*/;
	} else {
	    for (pos=0, s=theRuns->runs; pos+s->chars<=sp0.cp;	/* First ch */
		pos = pos+((s++)->chars)) /*loop*/;
	}

/*	Check our understanding */

	info = [self info:s];
	if (typingRun.paraStyle != 0) {
	    if  (typingRun.paraStyle != s->paraStyle)
	        printf("WWW: Strange: Typing run has bad style.\n");
	    if ((info != 0)
	        && ([self info:&typingRun] != info))
	        printf(
		"WWW: Strange: Typing run has bad anchor info.\n");
	}
	
	typingRun = *s;		/* Copy run to be used for insertion */
	run = *s;		/* save a copy */
    }
    
    if (!run.rFlags.dummy) return [super keyDown:theEvent]; // OK!

    {
	id result;
	[self prePaste];
	result = [super keyDown:theEvent];
	
/* 	Does it really change? YES!
*/    
	if (TRACE) {
	    if (typingRun.info != run.info) printf(
		"Typing run info was %p, now %p !!\n",
		(void*)run.info, (void*)typingRun.info);
	    if (typingRun.paraStyle != run.paraStyle) printf(
		"Typing run paraStyle was %p, now %p !!\n",
		run.paraStyle, typingRun.paraStyle);
	}
/*	Patch the new run if necessary:
*/
	{
	    int inserted = originalEnd-originalStart +
	    		 textLength-originalLength;
	    
	    if (TRACE) printf(
	"KeyDown, size(sel) %i (%i-%i)before, %i (%i-%i)after.\n",
		originalLength, originalStart, originalEnd,
		textLength, sp0.cp, spN.cp);

	    if (inserted>0) {
	    	NXRun * s;
		int pos;
		int start = sp0.cp-inserted;
    		for (pos=0, s=theRuns->runs; pos+s->chars<=start;
    			pos = pos+((s++)->chars)) /*loop*/;

//	s points to run containing first char of insertion

		if (pos!=start) {	/* insert in middle of run */
		    if (TRACE) printf(
			"HT: Inserted %i at %i, in run starting at=%i\n",
					inserted, start, pos);
					
		} else {	/* inserted stuff starts run */
		    if (TRACE) printf ("Patching info from %p to %p\n",
			    (void*)s->info, (void*)run.info);
	            [self setInfo:s to:info];
		    s->paraStyle = run.paraStyle;	/* free old one? */
		    s->rFlags.dummy = 1;
		}
	    } /* if inserted>0 */

	} /* block */
	return result;
    }
}
#endif

//	After paste, determine paragraph styles for pasted material:
//	------------------------------------------------------------

- (void) prePaste
{
    originalLength = textLength;
    originalStart = sp0.cp;
    originalEnd = spN.cp;

}

- (void) postPaste
{
    HTChildAnchor * typingInfo;
    int inserted = originalEnd-originalStart +
		    textLength-originalLength;
    
    if (TRACE) printf(
    "postPaste, size(sel) %i (%i-%i)before, %i (%i-%i)after.\n",
	originalLength, originalStart, originalEnd,
	textLength, sp0.cp, spN.cp);
	
    if (inserted>0) {
	NXRun *s, *r;
	NXTextStyle * prev_para;
	int pos;
	int start = sp0.cp-inserted;
	for (pos=0, s=theRuns->runs; pos+s->chars <= start;
		pos = pos+((s++)->chars)) /*loop*/;
//		s points to run containing first char of insertion

	if (pos!=sp0.cp-inserted)
		printf("HT paste: Strange: insert@%i != run@%i !!\n",
				start, pos);
				
	if (s > theRuns->runs) {
	    typingInfo = (HTChildAnchor *)((s-1)->info);
	    prev_para = s[-1].paraStyle;
	} else {
	    prev_para = (NXTextStyle*) HTStyleNamed(styleSheet, "");
	    typingInfo = 0;
	}
	for (r=s; pos+r->chars < sp0.cp; pos=pos+(r++)->chars) {
	    r->paraStyle = prev_para;
//	        // r->paraStyle = &HTStyleForRun(styleSheet, r)->paragraph;
	    [self setInfo:r to:(void*)typingInfo];
	}
	
    }
}


- paste:sender;
{
    id result;
    
    [self prePaste];
    
    result = [super paste:sender];		// Do the paste
        
    [self postPaste];
    
    return result;
}


@end

